#include <sp/spBaseLib.h>
#include <sp/spAudio.h>
#include <sp/spInputPlugin.h>

#include <sp/spMain.h>

static int debug_level;
static spBool help_flag;

static double samp_freq = 8000.0;
static int num_channel = 1;
static int samp_bit = 16;
static double rec_time = 3.0;
static spBool supported_only;
static long buffer_length = 0;
static int num_buffer;

static char *driver_name = NULL;
static char *plugin_search_path = NULL;

static spOptions options;
static spOption option[] = {
    {"-t", "-time", "recording time [sec]", NULL,
	 SP_TYPE_DOUBLE, &rec_time, "3.0"},
    {"-f", "-freq", "sampling frequency for raw data [Hz]", NULL,
	 SP_TYPE_DOUBLE, &samp_freq, "8000.0"},
    {"-c", "-channel", "number of channel for raw data", NULL,
	 SP_TYPE_INT, &num_channel, "1"},
    {"-b", "-bit", "bits per sample for raw data", NULL,
	 SP_TYPE_INT, &samp_bit, "16"},
    {"-so", NULL, "play supported file only", "supported_only",
	 SP_TYPE_BOOLEAN, &supported_only, SP_FALSE_STRING},
    {"-buf", NULL, "buffer length for audio output", "buffer_length",
	 SP_TYPE_LONG, &buffer_length, "1024"},
    {"-nbuf", NULL, "number of buffers (valid only on Windows)", "num_buffer",
	 SP_TYPE_LONG, &num_buffer, "16"},
    {"-driver", NULL, "audio driver name", "driver_name",
	 SP_TYPE_STRING, &driver_name, NULL},
    {"-path", NULL, "extra plugin search path", "plugin_path",
	 SP_TYPE_STRING, &plugin_search_path, NULL},
    {"-debug", NULL, "debug level", NULL,
	 SP_TYPE_INT, &debug_level, NULL},
    {"-h", "-help", "display this message", NULL,
	 SP_TYPE_BOOLEAN, &help_flag, SP_FALSE_STRING},
};

static char *filelabel[] = {
    "<output file>",
};

int spMain(int argc, char *argv[])
{
    long i;
    char *output_file;
    long buffer_size;
    double *data;
    long data_length;
    long total_length;
    long total_write;
    long nread, nwrite;
    spPlugin *o_plugin;
    spWaveInfo wave_info;
    spAudio audio;
	    
    /* get command line options */
    spSetHelpMessage(&help_flag, "test for playing sound file");
    options = spGetOptions(argc, argv, option, filelabel);
    spIgnoreUnknownOption(options, SP_TRUE);
    spGetOptionsValue(argc, argv, options);
    spSetDebugLevel(debug_level);
    spCheckNumFile(options);

    output_file = spGetFile(options);
    spDebug(1, "main", "output_file = %s\n", output_file);

    if ((audio = spInitAudioDriver(driver_name)) == NULL) {
	spError(1, "Can't initialize audio.\n");
    } else {
	spDebug(1, NULL, "init audio done\n");

	spInitWaveInfo(&wave_info);

	wave_info.samp_rate = samp_freq;
	wave_info.num_channel = num_channel;
	wave_info.samp_bit = samp_bit;
	
	buffer_size = buffer_length * (wave_info.samp_bit / 8);

	spSetAudioSampleRate(audio, wave_info.samp_rate);
	spSetAudioChannel(audio, wave_info.num_channel);
	spSetAudioSampleBit(audio, wave_info.samp_bit);
	spSetAudioNumBuffer(audio, num_buffer);
	spSetAudioBufferSize(audio, buffer_size);

	/* open output audio device */
	if (!spOpenAudioDevice(audio, "r")) {
	    spWarning("Can't open audio device.\n");
	} else {
	    spDebug(1, NULL, "open output audio device done\n");

	    total_length = (long)spRound(wave_info.samp_rate * rec_time) * wave_info.num_channel;
	    wave_info.length = total_length;
	    
	    if (supported_only) {
		o_plugin = spOpenFilePluginArg(NULL, output_file, "w", SP_PLUGIN_DEVICE_FILE,
					       &wave_info, NULL, argc, argv, NULL);
	    } else {
		o_plugin = spOpenFilePluginArgAuto("output_raw", output_file, "w", SP_PLUGIN_DEVICE_FILE,
						   &wave_info, NULL, argc, argv, NULL);
	    }

	    if (o_plugin != NULL) {
		/* create data */
		data_length = buffer_length;
		data = xalloc(data_length, double);

		for (i = 0, total_write = 0; total_write < total_length; i++) {
		    nread = spReadAudioDouble(audio, data, data_length);
		    
		    if ((nwrite = spWritePluginDouble(o_plugin, data, nread)) <= 0) {
			break;
		    }
		    spDebug(100, "main", "nwrite = %ld\n", nwrite);

		    total_write += nwrite;
		}

		xfree(data);
		
		spCloseFilePlugin(o_plugin);
	    } else {
		spWarning("Cannot open file: %s\n", output_file);
	    }
	    
	    spCloseAudioDevice(audio);
	}
	spFreeAudioDriver(audio);
    }
    
    return 0;
}
