/*
 *	plugin_alac.cpp
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spOutputPlugin.h>

#define SP_NUM_ALAC_TYPE 2
#define SP_ALAC_TYPE_MP4 0
#define SP_ALAC_TYPE_CAF 1

#define SP_ALAC_TYPE_DEFAULT SP_ALAC_TYPE_MP4

typedef struct _spAlacPluginInstance
{
    char type[SP_MAX_LINE];
    int type_index;
    
    FILE *fp;
    spLong64 current_pos;
    spLong64 total_duration;
    spSongInfoV2 song_info;

    spCafHeader cafHeader;
    
    spMp4Header mp4Header;
    spMp4TrackBox *mp4Trak;
    spMp4SampleTableBox *mp4Stbl;
    spMp4SampleToChunkBox *mp4Stsc;
    spMp4SampleEntryBox *mp4Soun;
    spMp4DataOnlyBox *mp4Alac;
    spMp4MetaBox *mp4Meta;
    unsigned long mp4DescriptionIndex;
    
    AudioFormatDescription format;
    
#if defined(SP_ALAC_INPUT_PLUGIN)
    ALACDecoder *decoder;
    BitBuffer inputBuffer;
    
    spFileOffset inputDataPos;
    spFileOffset prevInputDataPos;
#else
    AudioFormatDescription inputFormat;
    
    ALACEncoder *encoder;
    spLong64 totalDataBytes;
    spLong64 mp4ChunkOffset;
    unsigned long mp4SamplesPerChunk;
    spMp4Box *mp4PausedBox;
#endif

    unsigned char *readBuffer;
    unsigned char *writeBuffer;

    unsigned long bitsPerChannel;
    unsigned long bytesPerFrame;
    
    uint32_t numFrames;
    int32_t numBytes;
    
    unsigned long inputPacketBytes;
    unsigned long outputPacketBytes;
    unsigned long packetTableSize;
    unsigned long packetTableCount;
    
    unsigned long bufferCopyBytes;
} *spAlacPluginInstance;

static spBool spInitPluginAlac(const char *lang)
{
    /*spSetDebugLevel(10);*/

    return SP_TRUE;
}

static spBool spFreePluginAlac(void)
{

    return SP_TRUE;
}

static void *spInitPluginInstanceAlac(const char *lang)
{
    spAlacPluginInstance instance;
#if defined(SP_ALAC_INPUT_PLUGIN)
    ALACDecoder *theDecoder;

    theDecoder = new ALACDecoder;

    if (theDecoder == NULL) {
	spDebug(10, "spInitPluginInstanceAlac", "new ALACDecoder failed\n");
	return NULL;
    }
#else
    ALACEncoder *theEncoder;

    theEncoder = new ALACEncoder;

    if (theEncoder == NULL) {
	spDebug(10, "spInitPluginInstanceAlac", "new ALACEncoder failed\n");
	return NULL;
    }
#endif

    /*spSetDebugLevel(100);*/
    
    instance = xalloc(1, struct _spAlacPluginInstance);
    memset(instance, 0, sizeof(struct _spAlacPluginInstance));
    spStrCopy(instance->type, sizeof(instance->type), "");
    instance->type_index = SP_ALAC_TYPE_DEFAULT;
    
    instance->fp = NULL;
    instance->bitsPerChannel = 16;
    instance->current_pos = 0;
    instance->total_duration = 0;

    spInitSongInfoV2(&instance->song_info);
    
    instance->format.mSampleRate = 44100.0;
    instance->format.mChannelsPerFrame = 2;
    
#if defined(SP_ALAC_INPUT_PLUGIN)
    instance->decoder = theDecoder;
#else
    instance->encoder = theEncoder;
    instance->format.mFormatFlags = SP_CAF_ALAC_FORMAT_FLAG_16BIT_SOURCE_DATA;
#endif

    spDebug(50, "spInitPluginInstanceAlac", "done: instance = %ld\n", (long)instance);
    
    return (void *)instance;
}

static spBool spFreePluginInstanceAlac(void *instance)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;

    if (pinstance != NULL) {
#if defined(SP_ALAC_INPUT_PLUGIN)
	if (pinstance->decoder != NULL) {
	    delete pinstance->decoder;
	}
#else
	if (pinstance->encoder != NULL) {
	    delete pinstance->encoder;
	}
#endif
	
	xfree(pinstance);
    }
    
    return SP_TRUE;
}

static spBool spSetPluginFileTypeAlac(void *instance, int index)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    pinstance->type_index = index;
    
    return SP_TRUE;
}

static int spGetPluginFileTypeAlac(void *instance)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    return pinstance->type_index;
}

static spBool spSetPluginSongInfoAlac(void *instance, spSongInfo *song_info)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    spCopySongInfoV2(&pinstance->song_info, (spSongInfoV2 *)song_info);

    return SP_TRUE;
}

static spBool spGetPluginSongInfoAlac(void *instance, spSongInfo *song_info)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    spCopySongInfoV2((spSongInfoV2 *)song_info, &pinstance->song_info);

    return SP_TRUE;
}

static const char *spGetPluginBestSuffixAlac(void *instance)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    if (pinstance->type_index == SP_ALAC_TYPE_MP4) {
	return ".m4a";
    } else {
	return ".caf";
    }
}

static spBool spSetPluginSampleBitAlac(void *instance, int samp_bit)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    long formatFlags = -1;

    switch (samp_bit) {
      case 33:
      case 32:
	formatFlags = SP_CAF_ALAC_FORMAT_FLAG_32BIT_SOURCE_DATA;
	break;
      case 24:
	formatFlags = SP_CAF_ALAC_FORMAT_FLAG_24BIT_SOURCE_DATA;
	break;
      case 20:
	formatFlags = SP_CAF_ALAC_FORMAT_FLAG_20BIT_SOURCE_DATA;
	break;
      case 16:
	formatFlags = SP_CAF_ALAC_FORMAT_FLAG_16BIT_SOURCE_DATA;
	break;
      default:
	return SP_FALSE;
    }
    spDebug(50, "spSetPluginSampleBitAlac", "formatFlags = %ld, samp_bit = %d\n", formatFlags, samp_bit);

    pinstance->bitsPerChannel = samp_bit;
    pinstance->format.mFormatFlags = formatFlags;
    
    return SP_TRUE;
}

static spBool spGetPluginSampleBitAlac(void *instance, int *samp_bit)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;

    *samp_bit = pinstance->bitsPerChannel;
    
    return SP_TRUE;
}

static spBool spSetPluginChannelAlac(void *instance, int num_channel)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    pinstance->format.mChannelsPerFrame = num_channel;
    spDebug(50, "spSetPluginChannelAlac", "mChannelsPerFrame = %ld\n", pinstance->format.mChannelsPerFrame);
    
    return SP_TRUE;
}

static spBool spGetPluginChannelAlac(void *instance, int *num_channel)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    *num_channel = pinstance->format.mChannelsPerFrame;
    
    return SP_TRUE;
}

static spBool spSetPluginSampleRateAlac(void *instance, double samp_rate)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    pinstance->format.mSampleRate = samp_rate;
    spDebug(50, "spSetPluginSampleRateAlac", "mSampleRate = %f\n", pinstance->format.mSampleRate);
    
    return SP_TRUE;
}

static spBool spGetPluginSampleRateAlac(void *instance, double *samp_rate)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    *samp_rate = pinstance->format.mSampleRate;
    
    return SP_TRUE;
}

static spBool spSetPluginOtherInfoAlac(void *instance, const char *id, char *data)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    return SP_FALSE;
}

static spBool spGetPluginOtherInfoAlac(void *instance, const char *id, char *data)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    return SP_FALSE;
}
