/*
 *	input_flac.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spInputPluginP.h>
#include <sp/spPluginMain.h>

#include <FLAC/stream_decoder.h>
#include <FLAC/metadata.h>

#if !defined(SP_FLAC_INPUT_PLUGIN)
#define SP_FLAC_INPUT_PLUGIN 1
#endif

#include "plugin_flac.c"

static const char *sp_flac_file_type_list[] = {
    "flac",
    "oggflac",
    NULL,
};
static const char *sp_flac_file_desc_list[] = {
    "FLAC",
    "Ogg FLAC",
    NULL,
};
static const char *sp_flac_file_filter_list[] = {
    "*.flac",
    "*.ogg",
    NULL,
};
    
static spBool spIsSupportedByPluginFlac(const char *filename);

static void *spInitPluginInstanceFlac(const char *lang);
static spBool spFreePluginInstanceFlac(void *instance);

static spPluginError spOpenPluginFlac(void *instance, const char *filename, const char *mode);
static spBool spClosePluginFlac(void *instance);
static spPluginState spGetPluginStateFlac(void *instance);
static spLong spGetPluginCurrentPositionFlac(void *instance);
static spBool spStopPluginFlac(void *instance);
static spBool spPausePluginFlac(void *instance);
static spBool spRestartPluginFlac(void *instance);

static long spReadPluginFlac(void *instance, char *data, long length);
static spBool spSeekPluginFlac(void *instance, spLong pos);
static spLong spGetPluginTotalLengthFlac(void *instance);


static spInputPluginRec sp_input_plugin_flac = {
    NULL,
    NULL,

    SP_PLUGIN_INPUT,
    "FLAC",
    1,
    SP_PLUGIN_PRIORITY_HIGHEST,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginFlac,
    spFreePluginFlac,
    "Flac Input Plugin",
    "Flac Input Plugin Version 0.1\nCopyright(C) Hideki Banno\nBased on libFLAC (https://xiph.org/flac/api/index.html)",
    
    spInitPluginInstanceFlac,
    spFreePluginInstanceFlac,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_flac_file_type_list,
    sp_flac_file_desc_list,
    sp_flac_file_filter_list,
    spIsSupportedByPluginFlac,
    spSetPluginFileTypeFlac,  
    spGetPluginFileTypeFlac,
    (SP_SONG_TITLE_MASK | SP_SONG_ARTIST_MASK | SP_SONG_COMMENT_MASK
     | SP_SONG_GENRE_MASK | SP_SONG_RELEASE_MASK | SP_SONG_COPYRIGHT_MASK
     | SP_SONG_TRACK_MASK | SP_SONG_ISRC_MASK | SP_SONG_ALBUM_MASK | SP_SONG_SOFTWARE_MASK),
    spSetPluginSongInfoFlac,
    spGetPluginSongInfoFlac,
    spGetPluginBestSuffixFlac,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,

    spSetPluginSampleBitFlac,   
    spGetPluginSampleBitFlac,   
    spSetPluginChannelFlac,     
    spGetPluginChannelFlac,     
    spSetPluginSampleRateFlac,  
    spGetPluginSampleRateFlac,  
    spSetPluginOtherInfoFlac,   
    spGetPluginOtherInfoFlac,
    
    spOpenPluginFlac,              
    spClosePluginFlac,             
    spGetPluginStateFlac,          
    spGetPluginCurrentPositionFlac,
    spStopPluginFlac,              
    spPausePluginFlac,             
    spRestartPluginFlac,

    spReadPluginFlac,             
    spSeekPluginFlac,
    spGetPluginTotalLengthFlac,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_input_plugin_flac;
}

SP_DECLARE_UNIQUE_GET_PLUGIN_REC(input_flac)

static void *spInitPluginInstanceFlac(const char *lang)
{
    FLAC__StreamDecoder *decoder;
    spFlacPluginInstance instance;

    if ((decoder = FLAC__stream_decoder_new()) == NULL) {
	spDebug(1, "spInitPluginInstanceFlac", "ERROR: allocating decoder\n");
	return NULL;
    }
    
    instance = xalloc(1, struct _spFlacPluginInstance);
    instance->decoder = decoder;
    
    instance->decode_position = 0;
    instance->last_block_size = 0;
    instance->buffer_copy_pos = 0;

    instance->buffer = NULL;
    
    instance->total_samples = 0;
    instance->sample_rate = 0;
    instance->channels = 0;
    instance->bits_per_sample = 0;
    instance->max_blocksize = 0;

    instance->is_ogg = SP_FALSE;
    spInitSongInfoV2(&instance->song_info);
    
    instance->current_pos = 0;

    spDebug(10, "spInitPluginInstanceFlac", "done\n");
    
    return (void *)instance;
}

static spBool spFreePluginInstanceFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;

    if (pinstance == NULL) return SP_FALSE;

    if (pinstance->decoder != NULL) {
	FLAC__stream_decoder_delete(pinstance->decoder);
    }
    
    xfree(pinstance);
    
    return SP_TRUE;
}

static void getMetadataStreamInfo(spFlacPluginInstance pinstance, const FLAC__StreamMetadata *metadata)
{
    spDebug(10, "getMetadataStreamInfo", "metadata->type: FLAC__METADATA_TYPE_STREAMINFO\n");
    pinstance->total_samples = metadata->data.stream_info.total_samples;
    pinstance->sample_rate = metadata->data.stream_info.sample_rate;
    pinstance->channels = metadata->data.stream_info.channels;
    pinstance->bits_per_sample = metadata->data.stream_info.bits_per_sample;
    pinstance->total_size = pinstance->total_samples * pinstance->channels * (pinstance->bits_per_sample / 8);
    spDebug(10, "getMetadataStreamInfo", "metadata->data.stream_info.max_blocksize: %u\n", metadata->data.stream_info.max_blocksize);
    if (metadata->data.stream_info.max_blocksize == 0) {
	pinstance->max_blocksize = FLAC__MAX_BLOCK_SIZE;
    } else {
	pinstance->max_blocksize = metadata->data.stream_info.max_blocksize;
    }

    spDebug(10, "getMetadataStreamInfo", "sample rate    : %u Hz\n", pinstance->sample_rate);
    spDebug(10, "getMetadataStreamInfo", "channels       : %u\n", pinstance->channels);
    spDebug(10, "getMetadataStreamInfo", "bits per sample: %u\n", pinstance->bits_per_sample);
    spDebug(10, "getMetadataStreamInfo", "max blocksize: %u\n", pinstance->max_blocksize);
    spDebug(10, "getMetadataStreamInfo", "total samples  : %lu\n", (unsigned long)pinstance->total_samples);

    return;
}

static spBool getMetadataVorbisComment(spSongInfoV2 *song_info, const FLAC__StreamMetadata *metadata)
{
    unsigned int i, j;
    char *ostr;
    char *iptr, *optr;
    char *comment;
    char buf[SP_MAX_LINE];
    spConverter conv;
    unsigned long mask_done = 0L;

    spDebug(10, "getMetadataVorbisComment", "metadata->type: FLAC__METADATA_TYPE_VORBIS_COMMENT\n");

    if ((conv = spOpenConverter("UTF-8", NULL)) != NULL) {
	for (i = 0; i < metadata->data.vorbis_comment.num_comments; i++) {
	    comment = (char *)metadata->data.vorbis_comment.comments[i].entry;
	    spDebug(10, "getMetadataVorbisComment", "comments[%d] = %s\n", i, comment);
		
	    if (strvcaseeq(comment, "TRACKNUMBER=")) { /* track */
		song_info->info_mask |= SP_SONG_TRACK_MASK;
		song_info->track = atoi(comment + 12);
	    } else {
		for (j = 0; sp_flac_song_info_table[j].mask; j++) {
		    if (!(mask_done & sp_flac_song_info_table[j].mask)) {
			sprintf(buf, "%s=", sp_flac_song_info_table[j].name);
			spDebug(100, "getMetadataVorbisComment", "buf: %s, comment: %s\n", buf, comment);
			
			if (strvcaseeq(comment, buf)) {
			    mask_done |= sp_flac_song_info_table[j].mask;
			    song_info->info_mask |= sp_flac_song_info_table[j].mask;
			    optr = (char *)song_info + sp_flac_song_info_table[j].offset;
			    
			    iptr = comment + strlen(buf);

			    if ((ostr = xspConvert(conv, iptr)) != NULL) {
				spStrCopy(optr, sp_flac_song_info_table[j].size, ostr);
				xfree(ostr);
			    }
			}
		    }
		}
	    }
	}

#if 0
	spShowSongInfoV2(song_info, NULL, stdout);
#endif
	    
	spCloseConverter(conv);
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

static void metadataCallback(const FLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)client_data;

    if (pinstance == NULL) {
	return;
    }
    
    spDebug(10, "metadataCallback", "metadata->type = %d\n", metadata->type);

#if 0
    if (metadata->type == FLAC__METADATA_TYPE_STREAMINFO) {
	getMetadataStreamInfo(pinstance, metadata);
    }
#endif
}

static void errorCallback(const FLAC__StreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data)
{
    spDebug(1, "", "Got error callback: %s\n", FLAC__StreamDecoderErrorStatusString[status]);
}

static FLAC__StreamDecoderWriteStatus writeCallback(const FLAC__StreamDecoder *decoder,
						    const FLAC__Frame *frame,
						    const FLAC__int32 * const buffer[],
						    void *client_data)
{
    unsigned int c;
    spFlacPluginInstance pinstance = (spFlacPluginInstance)client_data;

    spDebug(80, "writeCallback", "in: blocksize = %u\n", frame->header.blocksize);
    
    if (pinstance != NULL) {
	if (frame->header.number_type == FLAC__FRAME_NUMBER_TYPE_SAMPLE_NUMBER) {
	    pinstance->decode_position = frame->header.number.sample_number;
	} else {
	    pinstance->decode_position = 0;
	}
	spDebug(80, "writeCallback", "current_pos = %ld, decode_position = %lu\n",
		(unsigned long)pinstance->current_pos, (unsigned long)pinstance->decode_position);
	
	for (c = 0; c < pinstance->channels; c++) {
	    memcpy(pinstance->buffer[c], buffer[c], frame->header.blocksize * sizeof(FLAC__int32));
	}
	pinstance->last_block_size = frame->header.blocksize;
	pinstance->buffer_copy_pos = 0;
    }

    spDebug(80, "writeCallback", "done\n");
    
    return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

static spBool spOpenFileFlac(FLAC__StreamDecoder *decoder, const char *filename, void *client_data, spBool is_ogg)
{
    FLAC__StreamDecoderInitStatus init_status;

    spDebug(80, "spOpenFileFlac", "filename = %s, is_ogg = %d\n", filename, is_ogg);
    
    FLAC__stream_decoder_set_md5_checking(decoder, true);
    
    /* open file */
    if (is_ogg) {
	init_status = FLAC__stream_decoder_init_ogg_file(decoder,
							 filename, writeCallback, metadataCallback,
							 errorCallback, client_data);
    } else {
	init_status = FLAC__stream_decoder_init_file(decoder,
						     filename, writeCallback, metadataCallback,
						     errorCallback, client_data);
    }
    
    spDebug(80, "spOpenFileFlac", "init_status = %d\n", init_status);
    
    if (init_status != FLAC__STREAM_DECODER_INIT_STATUS_OK) {
	spDebug(1, "spOpenFileFlac", "ERROR: initializing decoder: %s\n",
		FLAC__StreamDecoderInitStatusString[init_status]);
	return SP_FALSE;
    }
    
    return SP_TRUE;
}

static spBool spCloseFileFlac(FLAC__StreamDecoder *decoder)
{
    spBool flag = SP_TRUE;
    FLAC__StreamDecoderState state;

    state = FLAC__stream_decoder_get_state(decoder);
    spDebug(10, "spClosePluginFlac", "state = %d\n", state);
    
    if (state != FLAC__STREAM_DECODER_UNINITIALIZED
	&& state != FLAC__STREAM_DECODER_SEARCH_FOR_METADATA) {
	if (!FLAC__stream_decoder_finish(decoder)) {
	    spDebug(10, "spClosePluginFlac", "FLAC__stream_decoder_finish() FAILED\n");
	    flag = SP_FALSE;
	}
    }

    return flag;
}

static spBool spReadMetadataFlac(spFlacPluginInstance pinstance, const char *filename)
{
    spBool flag = SP_FALSE;
    FLAC__Metadata_Chain *chain;
    FLAC__Metadata_Iterator *it;
    FLAC__StreamMetadata *metadata;
    
    if ((chain = FLAC__metadata_chain_new()) != NULL) {
	if (FLAC__metadata_chain_read(chain, filename)
	    || FLAC__metadata_chain_read_ogg(chain, filename)) {
	    if ((it = FLAC__metadata_iterator_new()) != NULL) {
		FLAC__metadata_iterator_init(it, chain);

		do {
		    metadata = FLAC__metadata_iterator_get_block(it);
		    
		    if (metadata != NULL) {
			spDebug(10, "spReadMetadataFlac", "metadata->type = %d\n", metadata->type);
			if (metadata->type == FLAC__METADATA_TYPE_STREAMINFO) {
			    if (pinstance != NULL) {
				getMetadataStreamInfo(pinstance, (const FLAC__StreamMetadata *)metadata);
			    }
			    flag = SP_TRUE;
			} else if (metadata->type == FLAC__METADATA_TYPE_VORBIS_COMMENT) {
			    if (pinstance != NULL) {
				getMetadataVorbisComment(&pinstance->song_info, (const FLAC__StreamMetadata *)metadata);
			    }
			}
		    }
		} while (FLAC__metadata_iterator_next(it));
		
		FLAC__metadata_iterator_delete(it);
	    } else {
		spDebug(10, "spReadMetadataFlac", "FLAC__metadata_iterator_new failed: status = %d\n",
			FLAC__metadata_chain_status(chain));
	    }
	} else {
	    spDebug(10, "spReadMetadataFlac", "FLAC__metadata_chain_read failed: status = %d\n",
		    FLAC__metadata_chain_status(chain));
	}
	
	FLAC__metadata_chain_delete(chain);
    } else {
	spDebug(10, "spReadMetadataFlac", "FLAC__metadata_chain_new failed\n");
    }

    return flag;
}

static spBool spIsSupportedByPluginFlac(const char *filename)
{
    spBool flag;
    spBool is_ogg = SP_FALSE;
    FLAC__StreamDecoder *decoder;
    
    if (spCheckSuffixFlac(filename, &is_ogg) == SP_FALSE) {
	return SP_FALSE;
    }
    
    if (spReadMetadataFlac(NULL, filename) == SP_FALSE) {
	spDebug(1, "spIsSupportedByPluginFlac", "ERROR: spReadMetadataFlac failed\n");
	return SP_FALSE;
    }
    
    if ((decoder = FLAC__stream_decoder_new()) == NULL) {
	spDebug(1, "spIsSupportedByPluginFlac", "ERROR: allocating decoder\n");
	return SP_FALSE;
    }
    
    flag = spOpenFileFlac(decoder, filename, NULL, is_ogg);

    spCloseFileFlac(decoder);

    FLAC__stream_decoder_delete(decoder);
    
    return flag;
}

static spPluginError spOpenPluginFlac(void *instance, const char *filename, const char *mode)
{
    unsigned int c;
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    if (spCheckSuffixFlac(filename, &pinstance->is_ogg) == SP_FALSE) {
	return SP_PLUGIN_ERROR_OPEN;
    }
    
    pinstance->decode_position = 0;
    pinstance->total_samples = 0;
    pinstance->sample_rate = 0;
    pinstance->channels = 0;
    pinstance->bits_per_sample = 0;
    pinstance->max_blocksize = 0;
    
    pinstance->last_block_size = 0;
    pinstance->buffer_copy_pos = 0;

    if (spReadMetadataFlac(pinstance, filename) == SP_FALSE
	|| pinstance->sample_rate <= 0
	|| pinstance->channels <= 0
	|| pinstance->bits_per_sample <= 0
	|| pinstance->max_blocksize <= 0) {
	spDebug(1, "spOpenPluginFlac", "spReadMetadataFlac FAILED\n");
	return SP_PLUGIN_ERROR_OPEN;
    }

    /* open file */
    if (spOpenFileFlac(pinstance->decoder, filename, pinstance, pinstance->is_ogg) == SP_FALSE) {
	return SP_PLUGIN_ERROR_OPEN;
    }
    spDebug(10, "spOpenPluginFlac", "filename = %s\n", filename);

    if (!FLAC__stream_decoder_process_until_end_of_metadata(pinstance->decoder)) {
	spDebug(1, "spOpenPluginFlac", "FLAC__stream_decoder_process_until_end_of_metadata() FAILED\n");
	spClosePluginFlac(pinstance);
	return SP_PLUGIN_ERROR_OPEN;
    }

    pinstance->buffer = xalloc(pinstance->channels, FLAC__int32 *);
    for (c = 0; c < pinstance->channels; c++) {
	pinstance->buffer[c] = xalloc(pinstance->max_blocksize, FLAC__int32);
	memset(pinstance->buffer[c], 0, pinstance->max_blocksize * sizeof(FLAC__int32));
    }
    
    pinstance->current_pos = 0;

    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginFlac(void *instance)
{
    unsigned int c;
    spBool flag = SP_TRUE;
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;

    if (pinstance->buffer != NULL) {
	for (c = 0; c < pinstance->channels; c++) {
	    xfree(pinstance->buffer[c]);
	}
	xfree(pinstance->buffer);
	pinstance->buffer = NULL;
    }
    
    if (FLAC__stream_decoder_get_state(pinstance->decoder) != FLAC__STREAM_DECODER_UNINITIALIZED) {
	if (!FLAC__stream_decoder_finish(pinstance->decoder)) {
	    spDebug(10, "spClosePluginFlac", "FLAC__stream_decoder_finish() FAILED\n");
	    flag = SP_FALSE;
	}
    }
    
    pinstance->current_pos = 0;

    return flag;
}

static spPluginState spGetPluginStateFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    if (pinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static spLong spGetPluginCurrentPositionFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    return (spLong)pinstance->current_pos;
}

static spBool spStopPluginFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    FLAC__stream_decoder_reset(pinstance->decoder);
    pinstance->decode_position = 0;
    pinstance->current_pos = 0;
    pinstance->last_block_size = 0;
    pinstance->buffer_copy_pos = 0;


    return SP_TRUE;
}

static spBool spPausePluginFlac(void *instance)
{
    return SP_TRUE;
}

static spBool spRestartPluginFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    FLAC__stream_decoder_reset(pinstance->decoder);
    pinstance->decode_position = 0;
    pinstance->last_block_size = 0;
    pinstance->buffer_copy_pos = 0;

    return SP_TRUE;
}

static long spReadPluginFlac(void *instance, char *data, long length)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    unsigned int c;
    long k;
    long length2;
    long nread;
    short *sdata;
    long *ldata;
    FLAC__uint64 pos;
    
    spDebug(80, "spReadPluginFlac", "length = %ld, current_pos = %ld\n", length, pinstance->current_pos);
    spDebug(100, "spReadPluginFlac", "buffer_copy_pos = %u, last_block_size = %u\n",
	    pinstance->buffer_copy_pos, pinstance->last_block_size);

    sdata = (short *)data;
    ldata = (long *)data;

    pos = pinstance->current_pos;
    length2 = length / pinstance->channels;
    for (k = 0; k < length2; k++) {
	if (pinstance->last_block_size <= 0
	    || pinstance->buffer_copy_pos >= pinstance->last_block_size) {
	    pinstance->buffer_copy_pos = 0;
	    pinstance->last_block_size = 0;
	    if (!FLAC__stream_decoder_process_single(pinstance->decoder)
		|| pinstance->last_block_size <= 0) {
		if (FLAC__stream_decoder_get_state(pinstance->decoder) == FLAC__STREAM_DECODER_END_OF_STREAM) {
		    spDebug(1, "spReadPluginFlac", "last frame\n");
		    break;
		} else {
		    spDebug(1, "spReadPluginFlac", "FLAC__stream_decoder_process_single() FAILED\n");
		    return -1;
		}
	    }
	    spDebug(100, "spReadPluginFlac", "after FLAC__stream_decoder_process_single: buffer_copy_pos = %u, last_block_size = %u\n",
		    pinstance->buffer_copy_pos, pinstance->last_block_size);
	}
	
	for (c = 0; c < pinstance->channels; c++) {
	    if (pinstance->bits_per_sample <= 16) {
		sdata[k * pinstance->channels + c] = (short)pinstance->buffer[c][pinstance->buffer_copy_pos];
	    } else {
		ldata[k * pinstance->channels + c] = pinstance->buffer[c][pinstance->buffer_copy_pos];
	    }
	}
	pinstance->buffer_copy_pos++;
	pinstance->current_pos++;
    }

    nread = k;
    
    pos += k;
    spDebug(80, "spReadPluginFlac", "nread = %ld, pos = %ld, total_length = %ld\n",
	    nread, (long)pos, (long)pinstance->total_samples);

    if (pos > pinstance->total_samples) {
	nread = (long)MAX(pinstance->total_samples - (pos - k), 0);
	pos = pinstance->total_samples;
	spDebug(80, "spReadPluginFlac", "updated: nread = %ld, pos = %ld, total_length = %ld\n",
		nread, (long)pos, (long)pinstance->total_samples);
    }
    pinstance->current_pos = pos;
    
    return nread * pinstance->channels;
}

static spBool spSeekPluginFlac(void *instance, spLong pos)
{
    spBool flag = SP_TRUE;
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;

    spDebug(80, "spSeekPluginFlac", "pos = %ld\n", (long)pos);
    
    if (pos == pinstance->current_pos) {
	spDebug(80, "spSeekPluginFlac", "pos (%ld) is same as current_pos\n", (long)pos);
	return SP_TRUE;
    }

    pinstance->current_pos = pos;
    pinstance->last_block_size = 0;
    pinstance->buffer_copy_pos = 0;
    
    if (!FLAC__stream_decoder_seek_absolute(pinstance->decoder, pos)) {
	flag = SP_FALSE;
	if (FLAC__stream_decoder_get_state(pinstance->decoder) == FLAC__STREAM_DECODER_SEEK_ERROR) {
	    FLAC__stream_decoder_reset(pinstance->decoder);
	}
	pinstance->current_pos = 0;
    }

    
    return flag;
}

static spLong spGetPluginTotalLengthFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    return (spLong)pinstance->total_samples;
}

