/*
 *	plugin_flac.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spInputPluginP.h>

#define SP_NUM_FLAC_TYPE 2
#define SP_FLAC_TYPE_FLAC 0
#define SP_FLAC_TYPE_OGG 1

typedef struct _spFlacPluginInstance
{
#if defined(SP_FLAC_INPUT_PLUGIN)
    FLAC__StreamDecoder *decoder;
    
    FLAC__uint64 decode_position;
    
    unsigned last_block_size;
    unsigned buffer_copy_pos;
    
    FLAC__int32 **buffer;
#else
    FLAC__StreamEncoder *encoder;
    
    long encode_comp_level;
    unsigned encode_bits_per_sample;
    
    long encode_mult_weight;
    long encode_div_weight;

    FLAC__StreamMetadata *metadata;
    
    unsigned buffer_copy_pos;
    unsigned output_buffer_length;
    
    FLAC__int32 *output_buffer;
#endif
    
    FLAC__uint64 total_samples;
    unsigned sample_rate;
    unsigned channels;
    unsigned bits_per_sample;
    unsigned max_blocksize;
    FLAC__uint64 total_size;

    spBool is_ogg;
    int type_index;
    spSongInfoV2 song_info;
    
    FLAC__uint64 current_pos;
} *spFlacPluginInstance;

typedef struct _spFlacSongInfoTable {
    unsigned long mask;
    char *name;
    int offset;
    int size;
} spFlacSongInfoTable;

static spFlacSongInfoTable sp_flac_song_info_table[] = {
    {SP_SONG_ARTIST_MASK, "ARTIST", spOffset(spSongInfoV2Ptr, artist), SP_SONG_INFO_SIZE},
    {SP_SONG_ALBUM_MASK, "ALBUM", spOffset(spSongInfoV2Ptr, album), SP_SONG_INFO_SIZE},
    {SP_SONG_TITLE_MASK, "TITLE", spOffset(spSongInfoV2Ptr, title), SP_SONG_INFO_SIZE},
    {SP_SONG_GENRE_MASK, "GENRE", spOffset(spSongInfoV2Ptr, genre), SP_SONG_INFO_GENRE_SIZE},
    {SP_SONG_COPYRIGHT_MASK, "COPYRIGHT", spOffset(spSongInfoV2Ptr, copyright), SP_SONG_INFO_SIZE},
    {SP_SONG_RELEASE_MASK, "DATE", spOffset(spSongInfoV2Ptr, release), SP_SONG_INFO_RELEASE_SIZE},
    {SP_SONG_ISRC_MASK, "ISRC", spOffset(spSongInfoV2Ptr, isrc), SP_SONG_INFO_ISRC_SIZE},
    {SP_SONG_COMMENT_MASK, "DESCRIPTION", spOffset(spSongInfoV2Ptr, comment), SP_SONG_INFO_SIZE},
    {SP_SONG_SOFTWARE_MASK, "ENCODED BY", spOffset(spSongInfoV2Ptr, software), SP_SONG_INFO_SOFTWARE_SIZE},
    {SP_SONG_SOFTWARE_MASK, "ENCODER", spOffset(spSongInfoV2Ptr, software), SP_SONG_INFO_SOFTWARE_SIZE},
    {0, NULL, 0, 0},
};

static spBool spCheckSuffixFlac(const char *filename, spBool *is_ogg)
{
    if (spEqSuffix(filename, ".ogg") == SP_TRUE || spEqSuffix(filename, ".oga") == SP_TRUE) {
	*is_ogg = SP_TRUE;
	return SP_TRUE;
    } else if (spEqSuffix(filename, ".flac") == SP_TRUE) {
	*is_ogg = SP_FALSE;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spInitPluginFlac(const char *lang)
{
    /*spSetDebugLevel(20);*/
    return SP_TRUE;
}

static spBool spFreePluginFlac(void)
{
    return SP_TRUE;
}

static spBool spSetPluginFileTypeFlac(void *instance, int index)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    if (index == SP_FLAC_TYPE_OGG) {
	pinstance->is_ogg = SP_TRUE;
    } else if (index == SP_FLAC_TYPE_FLAC) {
	pinstance->is_ogg = SP_FALSE;
    } else {
	return SP_FALSE;
    }
    
    return SP_TRUE;
}

static int spGetPluginFileTypeFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    if (pinstance->is_ogg == SP_TRUE) {
	return SP_FLAC_TYPE_OGG;
    } else {
	return SP_FLAC_TYPE_FLAC;
    }
}

static spBool spSetPluginSongInfoFlac(void *instance, spSongInfo *song_info)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    spCopySongInfoV2(&pinstance->song_info, (spSongInfoV2 *)song_info);

    return SP_TRUE;
}

static spBool spGetPluginSongInfoFlac(void *instance, spSongInfo *song_info)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    spCopySongInfoV2((spSongInfoV2 *)song_info, &pinstance->song_info);

    return SP_TRUE;
}

static const char *spGetPluginBestSuffixFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    if (pinstance->is_ogg == SP_TRUE) {
	return ".ogg";
    } else {
	return ".flac";
    }
}

static spBool spSetPluginSampleBitFlac(void *instance, int samp_bit)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    if (samp_bit == 16 || samp_bit == 24 || samp_bit == 32) {
	pinstance->bits_per_sample = samp_bit;
#if !defined(SP_FLAC_INPUT_PLUGIN)
	pinstance->encode_bits_per_sample = samp_bit;
#endif
	spDebug(10, "spSetPluginSampleBitFlac", "bits_per_sample = %d\n", pinstance->bits_per_sample);
    } else {
	return SP_FALSE;
    }

    return SP_TRUE;
}

static spBool spGetPluginSampleBitFlac(void *instance, int *samp_bit)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;

    *samp_bit = pinstance->bits_per_sample;

    return SP_TRUE;
}

static spBool spSetPluginChannelFlac(void *instance, int num_channel)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    pinstance->channels = MAX(num_channel, 1);
    
    return SP_TRUE;
}

static spBool spGetPluginChannelFlac(void *instance, int *num_channel)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    *num_channel = pinstance->channels;

    return SP_TRUE;
}

static spBool spSetPluginSampleRateFlac(void *instance, double samp_rate)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    pinstance->sample_rate = (unsigned)spRound(samp_rate);

    return SP_TRUE;
}

static spBool spGetPluginSampleRateFlac(void *instance, double *samp_rate)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    *samp_rate = pinstance->sample_rate;

    return SP_TRUE;
}

static spBool spSetPluginOtherInfoFlac(void *instance, const char *id, char *data)
{
    
    return SP_FALSE;
}

static spBool spGetPluginOtherInfoFlac(void *instance, const char *id, char *data)
{

    return SP_FALSE;
}
